#IOCTL_READ_PORT_UCHAR	 = $200 
#IOCTL_WRITE_PORT_UCHAR	 = $400

CompilerIf Defined(PBIO_OK, #PB_Constant)=0
  Enumeration
    #PBIO_OK
    #PBIO_ErrCopyDriver
    #PBIO_ErrInstallDriver
    #PBIO_ErrAccess
  EndEnumeration
CompilerEndIf

Structure Param_Info
  hDriverID.i
  ErrStatus.i
  DrFileName.s
  DriverName.s
  *Point
  Len.i
EndStructure


Declare OpenDriver()
Declare Driver_UnInstall(Drivers_Name.s)

EnableExplicit

CompilerIf #PB_Compiler_Processor = #PB_Processor_x86

Prototype pIsWow64Process(hProcess.i, *Wow64Process)
Procedure OSbits()
  Protected Is64BitOS = 0 ; Returns 1 for 64bit, else 0.
  Protected hDLL, IsWow64Process
  
  Is64BitOS=0
  hDll = LoadLibrary_("kernel32.dll")
  If hDll
    
    IsWow64Process.pIsWow64Process = GetProcAddress_(hDll,"IsWow64Process")
    If IsWow64Process
      IsWow64Process(GetCurrentProcess_(), @Is64BitOS)
    EndIf
    
    FreeLibrary_(hDll)
  EndIf     
  
  ProcedureReturn Is64BitOS
EndProcedure
CompilerEndIf

ProcedureDLL IO_Init()
  
  Global ParamInfo.Param_Info
  
  ParamInfo\ErrStatus=#PBIO_OK
  CompilerIf #PB_Compiler_Processor = #PB_Processor_x86
    ParamInfo\DrFileName = GetTemporaryDirectory()+"pbdriverio"
    ParamInfo\DriverName = "pbdriverio"
    
    If OSbits() ; x64.
      ParamInfo\DrFileName+"_x64"
      ParamInfo\DriverName+"_x64"
      ParamInfo\Point = ?Driver_x64
      ParamInfo\Len   = ?DriverEnd_x64-?Driver_x64
    Else        ; x86.
      ParamInfo\Point = ?Driver
      ParamInfo\Len   = ?DriverEnd-?Driver
    EndIf
    ParamInfo\DrFileName+".sys"
  CompilerElse
    ParamInfo\DrFileName = GetTemporaryDirectory()+"pbdriverio_x64.sys"
    ParamInfo\DriverName = "pbdriverio_x64"
    ParamInfo\Point = ?Driver_x64
    ParamInfo\Len   = ?DriverEnd_x64-?Driver_x64
  CompilerEndIf
  
  ParamInfo\hDriverID = OpenDriver()
EndProcedure

ProcedureDLL IO_End()
  If ParamInfo\hDriverID
    CloseHandle_(ParamInfo\hDriverID)
    ParamInfo\hDriverID=0
  EndIf
EndProcedure


Procedure Driver_Start(Drivers_Name.s)
  Protected hSCManager.i, hServ.i
  Protected ErrorCode, Result
  
  Result=#False
  hSCManager  = OpenSCManager_(#Null, #Null, #SC_MANAGER_ALL_ACCESS)
  If hSCManager = 0
    If GetLastError_() = #ERROR_ACCESS_DENIED
      hSCManager  = OpenSCManager_(#Null, #Null, #GENERIC_READ)
      hServ       = OpenService_(hSCManager, Drivers_Name, #GENERIC_EXECUTE)
      If StartService_(hServ, 0, #Null)
        Result=#True
      EndIf
    EndIf
    
  Else
    hServ       = OpenService_(hSCManager, Drivers_Name, #GENERIC_EXECUTE)
    If StartService_(hServ, 0, #Null)
      Result=#True  
    Else
      ErrorCode=GetLastError_()
      If ErrorCode=1056
        Result=#True
      EndIf
    EndIf
  EndIf
  
  If hServ : CloseServiceHandle_(hServ) : EndIf
  If hSCManager : CloseServiceHandle_(hSCManager) : EndIf
  
  ProcedureReturn Result
EndProcedure


Procedure Driver_UnInstall(Drivers_Name.s)
  Protected hSCManager.i, hServ.i
  Protected ServiceStatus.SERVICE_STATUS
  
  hSCManager  = OpenSCManager_(#Null, #Null, #SC_MANAGER_ALL_ACCESS)
  hServ       = OpenService_(hSCManager, Drivers_Name, #SERVICE_ALL_ACCESS)
  
  If hServ<>#Null
     ControlService_(hServ, #SERVICE_CONTROL_STOP, @ServiceStatus)
     DeleteService_(hServ)
  EndIf
  
  If hServ : CloseServiceHandle_(hServ) : EndIf
  If hSCManager : CloseServiceHandle_(hSCManager) : EndIf
EndProcedure

Procedure Driver_Install(Drivers_Name.s)
  Protected Result, FileID, File.s, Mgr
  Protected ErrorCode, Ser, Count
  Result=#False
  Count=0 : FileID=0 : Mgr=0 : Ser=0

  File=ParamInfo\DrFileName
  
  
  If FileSize(File) <> ParamInfo\Len Or CRC32FileFingerprint(File) <> CRC32Fingerprint(ParamInfo\Point, ParamInfo\Len)
    
    FileID=CreateFile(#PB_Any, File)
    If FileID
      WriteData(FileID, ParamInfo\Point, ParamInfo\Len)
      CloseFile(FileID)
    ElseIf ParamInfo\ErrStatus = #PBIO_OK
      ParamInfo\ErrStatus = #PBIO_ErrCopyDriver
    EndIf
    
  Else
    FileID=2
  EndIf
  
  If FileID
    
    Driver_Install_M1:
    Mgr = OpenSCManager_(#Null, #Null, #SC_MANAGER_ALL_ACCESS)
    If Mgr
      Ser = CreateService_(Mgr, Drivers_Name, Drivers_Name, #SERVICE_ALL_ACCESS, #SERVICE_KERNEL_DRIVER, #SERVICE_DEMAND_START, #SERVICE_ERROR_NORMAL, File, #Null, 0, 0, 0, 0)
      ErrorCode=GetLastError_()
      If Ser 
        Result=#True 
      ElseIf ErrorCode=1073
        Result=#True 
      ElseIf ErrorCode=1072
        If Count=0
          
          If Ser : CloseServiceHandle_(Ser) : EndIf
          If Mgr : CloseServiceHandle_(Mgr) : EndIf
          Driver_UnInstall(Drivers_Name)
          Count+1
          Goto Driver_Install_M1
        EndIf
      EndIf
    EndIf
    
  EndIf
  
  If Ser : CloseServiceHandle_(Ser) : EndIf
  If Mgr : CloseServiceHandle_(Mgr) : EndIf
  
  If Result=#False And ParamInfo\ErrStatus = #PBIO_OK
    ParamInfo\ErrStatus = #PBIO_ErrInstallDriver
  EndIf
  
  ProcedureReturn Result
EndProcedure

Procedure OpenDriver()
  Protected Result, hDriver
  Result=0
  
  hDriver = CreateFile_("\\.\"+ParamInfo\DriverName, #GENERIC_READ | #GENERIC_WRITE, 0, 0, #OPEN_EXISTING, #FILE_ATTRIBUTE_NORMAL, 0)
  If hDriver=#INVALID_HANDLE_VALUE
    If Driver_Install(ParamInfo\DriverName)
      If Driver_Start(ParamInfo\DriverName)
        hDriver = CreateFile_("\\.\"+ParamInfo\DriverName, #GENERIC_READ | #GENERIC_WRITE, 0, 0, #OPEN_EXISTING, #FILE_ATTRIBUTE_NORMAL, 0)
        If hDriver <> #INVALID_HANDLE_VALUE
          Result=hDriver
        EndIf
      EndIf
    EndIf
  Else
    Result=hDriver
  EndIf
  
  If Result=0 And ParamInfo\ErrStatus = #PBIO_OK
    ParamInfo\ErrStatus = #PBIO_ErrAccess
  EndIf
  
  ProcedureReturn Result
EndProcedure


ProcedureDLL IO_Inp(PortAddress.u) ;    ,     'PortAddress'.
  Protected Result, Buffer.i, BytesReturned.i, OutBuff.i
  Result=0 : Buffer=0 : OutBuff=0
  
  If ParamInfo\hDriverID
    Buffer=PortAddress
     DeviceIoControl_(ParamInfo\hDriverID, #IOCTL_READ_PORT_UCHAR, @Buffer, 2, @OutBuff, 1, @BytesReturned, 0)
    Result=OutBuff
  EndIf
  
  ProcedureReturn Result
EndProcedure

ProcedureDLL IO_Out(PortAddress.u, Byte.a) ;    ,     'PortAddress'.
  Protected Result, Buffer.i, BytesReturned.i
  Result=0
  
  If ParamInfo\hDriverID
    Buffer=PortAddress
    PokeA(@Buffer+2, Byte)
   DeviceIoControl_(ParamInfo\hDriverID, #IOCTL_WRITE_PORT_UCHAR, @Buffer, 3, 0, 0, @BytesReturned, 0)
  EndIf
  
  ProcedureReturn Result
EndProcedure


ProcedureDLL IO_Status() ;  .   ,  0.
  ProcedureReturn ParamInfo\ErrStatus
EndProcedure


ProcedureDLL.f IO_Version()
  ProcedureReturn 1.4
EndProcedure



DataSection
  CompilerIf #PB_Compiler_Processor = #PB_Processor_x86
    Driver:
    IncludeBinary "Driver\pbdriverio.sys"
    DriverEnd:
  CompilerEndIf
  
  Driver_x64:
  IncludeBinary "Driver\pbdriverio_x64.sys"
  DriverEnd_x64:
EndDataSection
; IDE Options = PureBasic 5.00 (Windows - x86)
; ExecutableFormat = Shared Dll
; CursorPosition = 253
; FirstLine = 230
; Folding = --
; Executable = PB_IO.dll